# coding: utf-8
from django.db import transaction
from time import sleep


def nested_commit_on_success(func):
    """Like commit_on_success, but doesn't commit existing transactions.

    This decorator is used to run a function within the scope of a
    database transaction, committing the transaction on success and
    rolling it back if an exception occurs.

    Unlike the standard transaction.commit_on_success decorator, this
    version first checks whether a transaction is already active.  If so
    then it doesn't perform any commits or rollbacks, leaving that up to
    whoever is managing the active transaction.
    """
    commit_on_success = transaction.commit_on_success(func)

    def _nested_commit_on_success(*args, **kwds):
        if transaction.is_managed():
            return func(*args, **kwds)
        else:
            return commit_on_success(*args, **kwds)
    return transaction.wraps(func)(_nested_commit_on_success)


def dictfetchall(cursor):
    """Returns all rows from a cursor as a list of dicts"""
    desc = cursor.description
    return [
        dict(zip([col[0] for col in desc], row))
        for row in cursor.fetchall()
    ]


def dictfetchone(cursor):
    """Returns one row from a cursor as a dict"""
    desc = cursor.description
    return dict(zip([col[0] for col in desc], cursor.fetchone() or ()))


def smart_sleep(seconds, stop_event):
    """
    Приостановить выполнение на seconds секунд.
    При этом каждую секунду проверять, нет ли сигнала прерваться.
    :param stop_event: объект threading.Event. Нужен для остановки.
    """
    for count in xrange(seconds):
        sleep(1)
        if stop_event.is_set():
            break